from __future__ import annotations

import math


def expend_range(current_range, new_range):
    if current_range is None:
        return new_range

    return [
        min(current_range[0], new_range[0]),
        max(current_range[1], new_range[1]),
    ]


def to_precision(float_value, precision=3):
    precision_factor = math.pow(10, precision)
    int_value = int(float_value * precision_factor)
    return int_value / precision_factor



from __future__ import annotations

from trame.decorators import TrameApp, change
from trame.widgets import html
from trame.widgets import vuetify3 as v3
from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkDataSetAttributes

from parsli.utils import expend_range, to_precision
from parsli.viewer.vtk import PRESETS, set_preset, to_image


@TrameApp()
class ControlPanel(v3.VCard):
    def __init__(self, toggle, scene_manager, reset_camera, reset_to_mesh):
        self._scene_manager = scene_manager

        super().__init__(
            classes="controller",
            elevation=5,
            rounded=(f"{toggle} || 'circle'",),
        )

        # allocate variable if does not exist
        self.state.setdefault(toggle, True)
        self.state.setdefault("subdivide", True)
        self.state.setdefault("show_segment", True)
        self.state.setdefault("show_surface", True)
        self.state.setdefault("light_surface", False)
        self.state.setdefault("light_segment", False)

        with self:
            with v3.VCardTitle(
                classes=(
                    f"`d-flex align-center pa-1 position-fixed bg-white ${{ {toggle} ? 'controller-content rounded-t border-b-thin':'rounded-circle'}}`",
                ),
                style="z-index: 1;",
            ):
                v3.VProgressLinear(
                    v_if=toggle,
                    indeterminate=("trame__busy",),
                    bg_color="rgba(0,0,0,0)",
                    absolute=True,
                    color="primary",
                    location="bottom",
                    height=2,
                )
                v3.VProgressCircular(
                    v_else=True,
                    bg_color="rgba(0,0,0,0)",
                    indeterminate=("trame__busy",),
                    style="position: absolute; top: 0; left: 0; width: 100%; height: 100%;",
                    color="primary",
                    width=3,
                )
                v3.VBtn(
                    icon="mdi-close",
                    v_if=toggle,
                    click=f"{toggle} = !{toggle}",
                    flat=True,
                    size="sm",
                )
                v3.VBtn(
                    icon="mdi-menu",
                    v_else=True,
                    click=f"{toggle} = !{toggle}",
                    flat=True,
                    size="sm",
                )

                html.Div(
                    "Control Panel",
                    v_show=toggle,
                    classes="text-h6 px-2",
                )

                v3.VSpacer()

                v3.VProgressCircular(
                    "{{ export_progress }}",
                    v_if=("exporting_movie", False),
                    model_value=("export_progress", 0),
                    size=30,
                    classes="text-caption",
                )

                v3.VBtn(
                    v_else=True,
                    icon="mdi-movie-open-settings-outline",
                    density="compact",
                    flat=True,
                    click=self.ctrl.export_movie,
                    loading=("exporting_movie",),
                )

                # ------------------------------------------------
                # We now have control below, so removing for now
                # ------------------------------------------------
                # v3.VCheckbox(
                #     v_show=toggle,
                #     v_model=("show_segment", True),
                #     true_icon="mdi-gesture",
                #     false_icon="mdi-gesture",
                #     hide_details=True,
                #     density="compact",
                # )
                # v3.VCheckbox(
                #     v_show=toggle,
                #     v_model=("show_surface", True),
                #     true_icon="mdi-texture-box",
                #     false_icon="mdi-texture-box",
                #     hide_details=True,
                #     density="compact",
                #     classes="mx-2",
                # )
                # ------------------------------------------------

            with v3.VCardText(
                v_show=(toggle, True),
                classes="controller-content mt-12 mb-1 pb-1 mx-0 px-1",
            ):
                # -------------------------------------------------------------
                # Longitude / Latitude cropping
                # -------------------------------------------------------------

                with v3.VRow(
                    "Longitude", classes="text-subtitle-2 ma-1 pt-2 align-center"
                ):
                    v3.VSpacer()
                    html.Span(
                        "{{ longitude_bnds[0].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )
                    html.Span(
                        "{{ longitude_bnds[1].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )

                v3.VRangeSlider(
                    v_model=("longitude_bnds", [0, 360]),
                    min=0,
                    max=360,
                    step=0.5,
                    density="compact",
                    hide_details=True,
                    classes="px-2",
                )

                with v3.VRow("Latitude", classes="text-subtitle-2 ma-1 align-center"):
                    v3.VSpacer()
                    html.Span(
                        "{{ latitude_bnds[0].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )
                    html.Span(
                        "{{ latitude_bnds[1].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )

                v3.VRangeSlider(
                    v_model=("latitude_bnds", [-90, 90]),
                    min=-90,
                    max=90,
                    step=0.5,
                    density="compact",
                    hide_details=True,
                    classes="px-2",
                )

                with v3.VRow(classes="ma-1"):
                    v3.VBtn(
                        icon="mdi-crop-free",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        click=reset_camera,
                    )
                    v3.VBtn(
                        icon="mdi-magnify-scan",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=reset_to_mesh,
                    )

                    v3.VSpacer()
                    v3.VBtn(
                        icon="mdi-map-plus",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=self._expand_bounds,
                    )

                    v3.VBtn(
                        icon="mdi-arrow-collapse-horizontal",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        click=self._crop_bounds_to_mesh,
                    )

                    v3.VBtn(
                        icon="mdi-arrow-expand-horizontal",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=self._reset_bounds,
                    )

                # -------------------------------------------------------------

                v3.VDivider(classes="mt-2 mx-n3")

                # -------------------------------------------------------------
                # Projection: Spherical / Euclidean
                # -------------------------------------------------------------

                with html.Div(classes="d-flex"):
                    v3.VSelect(
                        prepend_inner_icon=(
                            "spherical ? 'mdi-earth' : 'mdi-earth-box'",
                        ),
                        v_model=("spherical", True),
                        items=(
                            "proj_modes",
                            [
                                {"title": "Spherical", "value": True},
                                {"title": "Euclidean", "value": False},
                            ],
                        ),
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                        classes="mx-n2",
                    )
                    v3.VCheckbox(
                        disabled=("!spherical",),
                        v_model=("show_earth_core", True),
                        true_icon="mdi-google-earth",
                        false_icon="mdi-google-earth",
                        hide_details=True,
                        density="compact",
                    )

                # -------------------------------------------------------------
                # Coast line regions
                # -------------------------------------------------------------

                v3.VSelect(
                    prepend_inner_icon="mdi-map-outline",
                    placeholder="Coast lines",
                    v_model=("coast_active_regions", []),
                    items=("coast_regions", []),
                    density="compact",
                    hide_details=True,
                    flat=True,
                    variant="solo",
                    chips=True,
                    closable_chips=True,
                    multiple=True,
                    classes="mx-n2",
                )
                # -------------------------------------------------------------

                v3.VDivider(classes="mx-n3 mb-1")

                # -------------------------------------------------------------
                # Opacity / Shadow
                # -------------------------------------------------------------

                v3.VSlider(
                    click_prepend="show_segment = !show_segment",
                    click_append="light_segment = !light_segment",
                    v_model=("segment_opacity", 100),
                    min=0,
                    max=1,
                    step=0.05,
                    prepend_icon="mdi-gesture",
                    append_icon=(
                        "light_segment ? 'mdi-lightbulb-outline' : 'mdi-lightbulb-off-outline'",
                    ),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-1",
                    style=("show_segment ? '' : 'opacity: 0.25'",),
                )

                v3.VSlider(
                    click_prepend="show_surface = !show_surface",
                    click_append="light_surface = !light_surface",
                    v_model=("surface_opacity", 100),
                    min=0,
                    max=1,
                    step=0.05,
                    prepend_icon="mdi-texture-box",
                    append_icon=(
                        "light_surface ? 'mdi-lightbulb-outline' : 'mdi-lightbulb-off-outline'",
                    ),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-1",
                    style=("show_surface ? '' : 'opacity: 0.25'",),
                )

                # -------------------------------------------------------------

                v3.VDivider(classes="mx-n3 mb-1")

                # -------------------------------------------------------------
                # Color mapping
                # -------------------------------------------------------------

                v3.VSelect(
                    placeholder="Color By",
                    prepend_inner_icon="mdi-format-color-fill",
                    v_model=("color_by", "dip_slip"),
                    items=("fields", []),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-n2",
                )

                with v3.VRow(no_gutters=True, classes="align-center mx-n2 mt-n2"):
                    with v3.VCol():
                        v3.VTextField(
                            v_model_number=("color_min", 0),
                            type="number",
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                            hide_spin_buttons=True,
                        )
                    with html.Div(classes="flex-0 mx-n3", style="z-index: 1;"):
                        v3.VBtn(
                            icon="mdi-arrow-expand-horizontal",
                            size="sm",
                            density="compact",
                            flat=True,
                            variant="solo",
                            classes="ml-2",
                            click=self.reset_color_range,
                        )
                        v3.VBtn(
                            icon="mdi-circle-half-full",
                            size="sm",
                            density="compact",
                            flat=True,
                            variant="solo",
                            classes="mr-2",
                            click=self.symetric_color_range,
                        )
                    with v3.VCol():
                        v3.VTextField(
                            v_model_number=("color_max", 1),
                            type="number",
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                            reverse=True,
                            classes="px-0",
                            hide_spin_buttons=True,
                        )

                html.Img(
                    src=("preset_img", None),
                    style="height: 1rem; width: 100%;",
                    classes="rounded-lg border-thin mb-n1",
                )

                v3.VSelect(
                    placeholder="Color Preset",
                    prepend_inner_icon="mdi-palette",
                    v_model=("color_preset", "Rycroft"),
                    items=("color_presets", list(PRESETS.keys())),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-n2",
                )

                v3.VDivider(classes="mx-n3 pb-1")

                # -------------------------------------------------------------
                # Contours
                # -------------------------------------------------------------

                with v3.VTooltip(
                    text=(
                        "`Number of contours: ${nb_contours} (${subdivide ? 'refined' : 'original'} mesh)`",
                    ),
                ):
                    with html.Template(v_slot_activator="{ props }"):
                        with html.Div(
                            classes="d-flex pr-2",
                            v_bind="props",
                        ):
                            v3.VSlider(
                                click_prepend="subdivide = !subdivide",
                                v_model=("nb_contours", 5),
                                min=1,
                                max=20,
                                step=1,
                                prepend_icon="mdi-fingerprint",
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                style="margin: 0 2px;",
                            )

                # -------------------------------------------------------------
                # Time
                # -------------------------------------------------------------

                with v3.VTooltip(
                    text=("`Current timestep: ${time_index + 1} / ${nb_timesteps}`",),
                ):
                    with html.Template(v_slot_activator="{ props }"):
                        with html.Div(
                            classes="d-flex pr-2",
                            v_bind="props",
                        ):
                            v3.VSlider(
                                v_model=("time_index", 0),
                                min=0,
                                max=("nb_timesteps - 1",),
                                step=1,
                                prepend_icon="mdi-clock-outline",
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                style="margin: 0 2px;",
                            )

    def _reset_bounds(self):
        self.state.latitude_bnds = [-90, 90]
        self.state.longitude_bnds = [0, 360]
        self.ctrl.view_update()

    def _expand_bounds(self):
        delta = 0.5
        self.state.latitude_bnds = [
            max(self.state.latitude_bnds[0] - delta, -90),
            min(self.state.latitude_bnds[1] + delta, 90),
        ]
        delta /= 2
        self.state.longitude_bnds = [
            max(self.state.longitude_bnds[0] - delta, 0),
            min(self.state.longitude_bnds[1] + delta, 360),
        ]
        self.ctrl.view_update()

    def _crop_bounds_to_mesh(self):
        source = self._scene_manager["meshes"].get("source")
        self.state.latitude_bnds = source.latitude_bounds
        self.state.longitude_bnds = source.longitude_bounds
        self.ctrl.view_update()

    @change("time_index")
    def _time_change(self, time_index, **_):
        source = self._scene_manager["meshes"].get("source")
        source.time_index = time_index
        self.ctrl.view_update()

    @change("coast_active_regions")
    def _on_regions(self, coast_active_regions, **_):
        source = self._scene_manager["coast"].get("source")
        source.active_regions = coast_active_regions
        self.ctrl.view_update()

    @change(
        "show_segment",
        "show_surface",
        "show_earth_core",
        "light_segment",
        "light_surface",
        "surface_opacity",
        "segment_opacity",
    )
    def _on_visibility(
        self,
        show_segment,
        show_surface,
        show_earth_core,
        light_segment,
        light_surface,
        surface_opacity,
        segment_opacity,
        **_,
    ):
        seg_actors = self._scene_manager["segment"].get("actors")
        surf_actors = self._scene_manager["meshes"].get("actors")
        earth_actors = self._scene_manager["earth_core"].get("actors")

        for actor in seg_actors:
            actor.SetVisibility(show_segment)
            actor.property.opacity = segment_opacity
            actor.property.lighting = not light_segment

        for actor in surf_actors:
            actor.SetVisibility(show_surface)
            actor.property.opacity = surface_opacity
            actor.property.lighting = not light_surface

        for actor in earth_actors:
            actor.SetVisibility(show_earth_core)

        self.ctrl.view_update()

    @change("latitude_bnds", "longitude_bnds")
    def _on_lat_lon_bnd(self, longitude_bnds, latitude_bnds, **_):
        if self._scene_manager["segment"]:
            reader = self._scene_manager["segment"].get("source")
            reader.longitude_bnds = longitude_bnds
            reader.latitude_bnds = latitude_bnds
        self.ctrl.view_update()

    @change("color_by", "color_preset", "color_min", "color_max", "nb_contours")
    def _on_color_preset(
        self, color_preset, color_min, color_max, color_by, nb_contours, **_
    ):
        lut = self._scene_manager.lut
        color_min = float(color_min)
        color_max = float(color_max)

        for mesh_type in ["segment", "meshes"]:
            if self._scene_manager[mesh_type]:
                mapper = self._scene_manager[mesh_type].get("mapper")
                mapper.SetScalarVisibility(1)

                if mesh_type == "segment":
                    mapper.SelectColorArray(color_by)
                    mapper.SetScalarRange(color_min, color_max)
                else:
                    bands = self._scene_manager[mesh_type].get("bands")

                    # The scalar we color on is an int starting
                    # at 0 for the first band.
                    # => Sample color at the center of the band by 0.5 shift
                    mapper.SetScalarRange(-0.5, nb_contours + 0.5)

        mesh_pipeline = self._scene_manager["meshes"]
        assign = mesh_pipeline.get("assign")
        assign.Assign(
            color_by,
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        bands = mesh_pipeline.get("bands")

        # +2 because for 1 cut line we need 3 values [min, cut_line, max]
        bands.GenerateValues(nb_contours + 2, [color_min, color_max])
        bands.ClippingOn()

        if "color_preset" in self.state.modified_keys:
            set_preset(lut, color_preset)
            self.state.preset_img = to_image(lut, 255)

        self.ctrl.view_update()

    def symetric_color_range(self):
        max_bound = max(abs(self.state.color_min), abs(self.state.color_max))
        max_bound = to_precision(max_bound, 3)
        self.state.color_min = -max_bound
        self.state.color_max = max_bound

    def reset_color_range(self):
        pipeline_item = self._scene_manager["meshes"]
        source = pipeline_item.get("source")
        ds = source()

        total_range = None
        for array in ds.cell_data[self.state.color_by].Arrays:
            total_range = expend_range(total_range, array.GetRange())

        self.state.color_min = to_precision(total_range[0], 3)
        self.state.color_max = to_precision(total_range[1], 3)


class ViewToolbar(v3.VCard):
    def __init__(self, reset_camera, reset_to_mesh, apply_zoom, update_view_up):
        super().__init__(
            classes="view-toolbar pa-1",
            rounded="lg",
        )

        self.state.setdefault("interaction_style", "trackball")

        with self:
            with v3.VTooltip(text="Reset camera"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-crop-free",
                        click=reset_camera,
                    )

            with v3.VTooltip(text="Reset camera centered on mesh"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-scan",
                        click=reset_to_mesh,
                    )
            v3.VDivider()
            with v3.VTooltip(text="Zoom in"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-plus-outline",
                        click=(apply_zoom, "[1.1]"),
                    )
            with v3.VTooltip(text="Zoom out"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-minus-outline",
                        click=(apply_zoom, "[0.9]"),
                    )

            v3.VDivider()
            with v3.VTooltip(text="Trackball interaction"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_show="interaction_style == 'trackball'",
                        disabled=("!spherical",),
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-rotate-orbit",
                        click="interaction_style = 'terrain'",
                    )

            with v3.VTooltip(text="Terrain interaction"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_show="interaction_style == 'terrain'",
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-axis-z-rotate-counterclockwise",
                        click="interaction_style = 'trackball'",
                    )

            with v3.VTooltip(text="Z up"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_show="interaction_style == 'trackball' && spherical",
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-axis-z-arrow",
                        click=(update_view_up, "[[0, 0, 1]]"),
                    )
            with v3.VTooltip(text="Y up"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_show="!spherical",
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-axis-y-arrow",
                        click=(update_view_up, "[[0, 1, 0]]"),
                    )


class ScalarBar(v3.VTooltip):
    def __init__(
        self,
        img_src="preset_img",
        color_min="color_min",
        color_max="color_max",
        **kwargs,
    ):
        super().__init__(location="top")

        self.state.setdefault("scalarbar_probe", [])
        self.state.client_only("scalarbar_probe", "scalarbar_probe_available")

        with self:
            # Content
            with html.Template(v_slot_activator="{ props }"):
                with html.Div(
                    classes="scalarbar",
                    rounded="pill",
                    v_bind="props",
                    **kwargs,
                ):
                    html.Div(f"{{{{ {color_min} }}}}", classes="scalarbar-left")
                    html.Img(
                        src=(img_src, None),
                        style="height: 100%; width: 100%;",
                        classes="rounded-lg border-thin",
                        mousemove="scalarbar_probe = [$event.x, $event.target.getBoundingClientRect()]",
                        mouseenter="scalarbar_probe_available = 1",
                        mouseleave="scalarbar_probe_available = 0",
                        __events=["mousemove", "mouseenter", "mouseleave"],
                    )
                    html.Div(
                        v_show=("scalarbar_probe_available", False),
                        classes="scalar-cursor",
                        style=(
                            "`left: ${scalarbar_probe?.[0] - scalarbar_probe?.[1]?.left}px`",
                        ),
                    )
                    html.Div(f"{{{{ {color_max} }}}}", classes="scalarbar-right")
            html.Span(
                f"{{{{ (({color_max} - {color_min}) * (scalarbar_probe?.[0] - scalarbar_probe?.[1]?.left) / scalarbar_probe?.[1]?.width + {color_min}).toFixed(3) }}}}"
            )



from __future__ import annotations

import asyncio
import time
from datetime import date
from pathlib import Path

from trame.app import asynchronous, get_server
from trame.decorators import TrameApp, change, controller
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import vtk as vtkw
from trame.widgets import vtklocal
from trame.widgets import vuetify3 as v3
from vtkmodules.vtkFiltersSources import vtkSphereSource
from vtkmodules.vtkIOParallelXML import vtkXMLPartitionedDataSetWriter

from parsli.io import VtkCoastLineSource, VtkMeshReader, VtkSegmentReader
from parsli.utils import expend_range, to_precision
from parsli.utils.earth import EARTH_RADIUS
from parsli.viewer import css, ui
from parsli.viewer.vtk import SceneManager

DEBUG_WRITE_MESH = False


@TrameApp()
class Viewer:
    def __init__(self, server=None):
        self.server = get_server(server, client_type="vue3")
        self.server.enable_module(css)
        self.server.cli.add_argument(
            "--data", help="Path of hdf5 file to load", required=True
        )
        self.server.cli.add_argument(
            "--wasm", help="Use local rendering", action="store_true"
        )

        # process cli
        args, _ = self.server.cli.parse_known_args()
        data_file = str(Path(args.data).resolve())
        self.local_rendering = args.wasm

        # Setup app
        self.scene_manager = SceneManager(self.server)
        self._build_ui()

        # earth core
        pipeline = self.scene_manager.add_geometry(
            "earth_core",
            vtkSphereSource(
                radius=EARTH_RADIUS - 100,
                theta_resolution=60,
                phi_resolution=60,
            ),
        )
        prop = pipeline.get("actor").property
        prop.opacity = 0.85

        # load segments
        seg_reader = VtkSegmentReader()
        seg_reader.file_name = data_file
        pipeline = self.scene_manager.add_geometry("segment", seg_reader)
        pipeline.get("mapper").SetScalarModeToUseCellFieldData()

        # load meshes
        mesh_reader = VtkMeshReader()
        mesh_reader.file_name = data_file
        pipeline = self.scene_manager.add_geometry_with_contour(
            "meshes", mesh_reader, True
        )
        pipeline.get(
            "mapper"
        ).SetScalarModeToUseCellFieldData()  # Bands: Scalars on Cell
        self.state.fields = mesh_reader.available_fields
        self.state.time_index = mesh_reader.time_index
        self.state.nb_timesteps = mesh_reader.number_of_timesteps

        # Coast lines
        self.coast_lines = VtkCoastLineSource()
        self.state.coast_regions = self.coast_lines.available_regions
        self.state.coast_active_regions = []
        pipeline = self.scene_manager.add_geometry("coast", self.coast_lines, True)
        coast_props = pipeline.get("actor").property
        coast_props.line_width = 2
        coast_props.color = (0, 0, 0)

        if DEBUG_WRITE_MESH:
            writer = vtkXMLPartitionedDataSetWriter()
            writer.SetInputData(mesh_reader())
            writer.SetFileName("all_meshes.vtpd")
            writer.Write()

        self.readers = [mesh_reader, seg_reader]

        # setup camera to look at the data
        bounds = self.scene_manager["meshes"].get("actor").bounds
        self.scene_manager.focus_on(bounds)

    @property
    def ctrl(self):
        return self.server.controller

    @property
    def state(self):
        return self.server.state

    @change("color_by")
    def _on_color_by(self, color_by, **_):
        pipeline_item = self.scene_manager["meshes"]
        source = pipeline_item.get("source")
        mapper_mesh = pipeline_item.get("mapper")
        mapper_seg = self.scene_manager["segment"].get("mapper")

        if color_by is None:
            mapper_mesh.SetScalarVisibility(0)
            mapper_seg.SetScalarVisibility(0)
            self.ctrl.view_update()
            return

        # Extract data range
        ds = source()

        total_range = None
        for array in ds.cell_data[color_by].Arrays:
            total_range = expend_range(total_range, array.GetRange())

        self.state.color_min = to_precision(total_range[0], 3)
        self.state.color_max = to_precision(total_range[1], 3)

    @change("spherical")
    def _on_projection_change(self, spherical, **_):
        self.state.show_earth_core = spherical

        for geo_name in ["segment", "meshes", "coast"]:
            pipeline_item = self.scene_manager[geo_name]
            pipeline_item.get("source").spherical = spherical
            actors = pipeline_item.get("actors")

            scale = (1, 1, 1) if spherical else (1, 1, 0.01)
            for actor in actors:
                actor.scale = scale

        if spherical:
            bounds = self.scene_manager["meshes"].get("actor").bounds
            self.scene_manager.camera.focal_point = (0, 0, 0)
            self.scene_manager.camera.view_up = (0, 0, 1)
            self.scene_manager.focus_on(bounds)
        else:
            self.state.interaction_style = "trackball"
            self.scene_manager.camera.focal_point = (0, 0, 0)
            self.scene_manager.camera.position = (0, 0, 1)
            self.scene_manager.camera.view_up = (0, 1, 0)

        self.reset_to_mesh()

    @change("camera")
    def _on_camera(self, camera, **_):
        if camera is None:
            return

        self.ctrl.vtk_update_from_state(camera)

    @change("interaction_style")
    def _on_style_change(self, interaction_style, **_):
        self.scene_manager.update_interaction_style(interaction_style)
        self.ctrl.view_update(push_camera=True)

    @change("subdivide")
    def _on_subdivide(self, subdivide, **_):
        # source >> quality >> threshold >> geometry >> cell2point >> refine >> assign >> bands
        pipeline = self.scene_manager["meshes"]

        source = pipeline.get("source")
        assign = pipeline.get("assign")
        refine = pipeline.get("refine")
        cell2point = pipeline.get("cell2point")
        threshold = pipeline.get("threshold")
        geometry = pipeline.get("geometry")

        if subdivide:
            geometry.input_connection = threshold.output_port
            assign.input_connection = refine.output_port

            # debug
            # self.debug_check_quality()
        else:
            geometry.input_connection = source.output_port
            assign.input_connection = cell2point.output_port

        self.ctrl.view_update()

    def reset_to_mesh(self):
        bounds = self.scene_manager["meshes"].get("actor").bounds
        self.scene_manager.reset_camera_to(bounds)
        self.ctrl.view_update(push_camera=True)

    def apply_zoom(self, scale):
        self.scene_manager.apply_zoom(scale)
        self.ctrl.view_update(push_camera=True)

    def update_view_up(self, view_up):
        self.scene_manager.update_view_up(view_up)
        self.ctrl.view_update(push_camera=True)

    # def debug_check_quality(self):
    #     filter = self.scene_manager["meshes"].get("quality").GetOutputDataObject(0)
    #     for array in filter.cell_data["Quality"].Arrays:
    #         print("Quality range", array.GetRange())

    async def _export_movie(self):
        t0 = time.time()
        await asyncio.sleep(0.1)

        # find export location
        base = f"export-{date.today().isoformat()}"
        current_directory_name = base
        count = 1
        while Path(current_directory_name).exists():
            current_directory_name = f"{base}-{count}"
            count += 1

        base_directory = Path(current_directory_name)
        base_directory.mkdir(parents=True)
        print(  # noqa: T201
            "\n----------------------------------------"
            "\nExporting images:"
            f"\n => location: {base_directory.resolve()}"
            f"\n => number of frames: {self.state.nb_timesteps}"
        )

        source = self.scene_manager["meshes"].get("source")
        futures = []
        nb_timesteps = self.state.nb_timesteps
        for t_idx in range(nb_timesteps):
            source.time_index = t_idx % self.state.nb_timesteps
            futures.append(
                self.scene_manager.write_jpg_screenshot(
                    base_directory / f"{t_idx:012}.jpg"
                )
            )
            progress = int(100 * (t_idx + 1) / nb_timesteps)
            if progress != self.state.export_progress:
                with self.state:
                    self.state.export_progress = progress
                await asyncio.sleep(0.001)

        # Ensure full completion
        for future in futures:
            future.result()

        t2 = time.time()
        print(f" => time: {t2 - t0:.1f}s")  # noqa: T201
        print(f" => fps: {nb_timesteps / (t2 - t0):.1f}")  # noqa: T201

        with self.state:
            self.state.exporting_movie = False
            self.state.export_progress = 100

    @controller.set("export_movie")
    def export_movie(self):
        self.state.export_progress = 0
        self.state.exporting_movie = True
        asynchronous.create_task(self._export_movie())

    def _build_ui(self):
        self.state.trame__title = "Parsli"
        self.state.setdefault("camera", None)
        with VAppLayout(self.server, full_height=True) as layout:
            self.ui = layout

            with v3.VContainer(
                fluid=True, classes="fill-height pa-0 ma-0 position-relative"
            ):
                if self.local_rendering:
                    with vtklocal.LocalView(
                        self.scene_manager.render_window,
                        20,
                        camera="camera = $event",
                    ) as view:
                        view.register_vtk_object(self.scene_manager.widget)
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera
                        self.ctrl.vtk_update_from_state = view.vtk_update_from_state
                else:
                    with vtkw.VtkRemoteView(
                        self.scene_manager.render_window,
                        interactive_ratio=2,
                        still_ratio=2,
                    ) as view:
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera

                # Control panel
                ui.ControlPanel(
                    toggle="show_panel",
                    scene_manager=self.scene_manager,
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                )

                # 3D View controls
                ui.ViewToolbar(
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                    apply_zoom=self.apply_zoom,
                    update_view_up=self.update_view_up,
                )

                # ScalarBar
                ui.ScalarBar()
